use forgejo_api::Forgejo;
use std::collections::HashMap;
use tokio::time::{sleep, Duration};

use crate::data::*;

async fn scrape_repos(forge: &Forgejo) -> anyhow::Result<Vec<forgejo_api::structs::Repository>> {
    let mut repos = Vec::new();
    let mut query = forgejo_api::structs::RepoSearchQuery {
        limit: Some(50),
        ..Default::default()
    };
    let mut page: u32 = 1;
    loop {
        query.page = Some(page);
        let resp = forge.repo_search(query.clone()).await?;
        match (resp.ok, resp.data) {
            (Some(true), Some(mut query_repos)) => {
                if query_repos.is_empty() {
                    break;
                }
                repos.append(&mut query_repos);
            }
            _ => todo!("scrape_repos: implement retries"),
        }
        page += 1;
        sleep(Duration::from_millis(100)).await;
    }
    Ok(repos)
}

async fn scrape_issues(forge: &Forgejo) -> anyhow::Result<Vec<forgejo_api::structs::Issue>> {
    let mut issues = Vec::new();
    let mut query = forgejo_api::structs::IssueSearchIssuesQuery {
        limit: Some(50),
        ..Default::default()
    };
    let mut page: u32 = 1;
    loop {
        query.page = Some(page);
        let mut resp = forge.issue_search_issues(query.clone()).await?;
        if resp.is_empty() {
            break;
        }
        issues.append(&mut resp);
        page += 1;
        sleep(Duration::from_millis(100)).await;
    }
    Ok(issues)
}

async fn scrape_users(forge: &Forgejo) -> anyhow::Result<Vec<forgejo_api::structs::User>> {
    let mut users = Vec::new();
    let mut query = forgejo_api::structs::UserSearchQuery {
        limit: Some(50),
        ..Default::default()
    };
    let mut page: u32 = 1;
    loop {
        query.page = Some(page);
        let resp = forge.user_search(query.clone()).await?;
        match (resp.ok, resp.data) {
            (Some(true), Some(mut query_repos)) => {
                if query_repos.is_empty() {
                    break;
                }
                users.append(&mut query_repos);
            }
            _ => todo!("scrape_users: implement retries"),
        }
        page += 1;
        sleep(Duration::from_millis(20)).await;
    }
    Ok(users)
}

pub async fn get_user_data(forge: &Forgejo) -> anyhow::Result<HashMap<UserId, UserData>> {
    let mut data = HashMap::new();

    let discard_empty = |o: Option<String>| match o {
        None => None,
        Some(s) if s.is_empty() => None,
        Some(s) => Some(s),
    };

    eprintln!("Fetching users...");
    for user in scrape_users(forge).await? {
        let Some(id) = user.id else {
            eprintln!("WARN: user with no id");
            continue;
        };
        let Some(login) = user.login else {
            eprintln!("WARN: missing login for user {id}");
            continue;
        };

        let Some(email) = user.email else {
            eprintln!("WARN: missing email for user {id}");
            continue;
        };

        data.insert(
            UserId(id),
            UserData {
                login,
                email,
                full_name: discard_empty(user.full_name),
                location: discard_empty(user.location),
                website: discard_empty(user.website),
                description: discard_empty(user.description),
                repos: Vec::new(),
                issues: Vec::new(),
            },
        );
    }

    eprintln!("Fetching repos...");
    for repo in scrape_repos(forge).await? {
        let Some(id) = repo.id else {
            eprintln!("WARN: repo with no id");
            continue;
        };
        let Some(owner) = repo.owner else {
            eprintln!("WARN: repo {} with no owner", id);
            continue;
        };
        let Some(owner_id) = owner.id else {
            eprintln!("WARN: owner for repo {} has no id", id);
            continue;
        };
        let Some(repo_name) = repo.name else {
            eprintln!("WARN: repo {} has no name", id);
            continue;
        };
        let Some(forge_owner) = data.get_mut(&UserId(owner_id)) else {
            // this currently happens for repos owned by organizations
            eprintln!(
                "WARN: repo owner {} for repo {} is not in database",
                owner.login.unwrap_or_default(),
                repo_name
            );
            continue;
        };
        forge_owner.repos.push((
            RepoId(id),
            RepoData {
                name: repo_name,
                description: discard_empty(repo.description),
            },
        ));
    }

    eprintln!("Fetching issues...");
    for issue in scrape_issues(forge).await? {
        let Some(id) = issue.id else {
            eprintln!("WARN: issue with no id");
            continue;
        };
        let Some(user) = issue.user else {
            eprintln!("WARN: issue {} has no owner", id);
            continue;
        };
        let Some(user_id) = user.id else {
            eprintln!("WARN: user for issue {} has no id", id);
            continue;
        };
        let Some(forge_user) = data.get_mut(&UserId(user_id)) else {
            eprintln!(
                "WARN: issue user {} {} for issue {} is not in database",
                user.login.unwrap_or_default(),
                user_id,
                issue
                    .html_url
                    .map_or(String::from(""), |url| url.as_str().to_string())
            );
            continue;
        };
        forge_user.issues.push((
            IssueId(id),
            IssueData {
                title: issue.title.unwrap_or_default(),
                body: issue.body.unwrap_or_default(),
            },
        ));
    }

    // discard users with an entirely empty profile: there is nothing useful we
    // can say about them
    let data = data
        .into_iter()
        .filter(|(_, user)| !user.is_empty())
        .collect();
    Ok(data)
}
